/*
 *	Copyright (c) 1994, 1995 by Jeremy Cooperstock and Steve Kotsopoulos,
 *	All Rights Reserved.
 */

/*
 * afdpsend.c - an AFDP client (sends files to a group)
 *
 * static char rcsid[] = "$Id: afdpsend.c,v 1.97 1998/07/02 19:37:36 steve Exp $";
 */

#ifdef NEWSGI
#include <getopt.h>
#endif
#include "config.h"
#include "sockutils.h"
#include "afdp.h"
#include "afdplib.h"
#include "afdpsend.h"
#include <time.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#ifndef MIPS
#include <unistd.h>
#endif

/* general global variables */
int	debug = 0;		/* global debugging variable		*/
group_id_t group = 1;		/* the group we are going to send to	*/
char	*command = NULL; 	/* command for subscriber to run	*/
char *progname;
char *portname = "afdp";
boolean_t daemonize = FALSE;
boolean_t reserved_ports = FALSE;
boolean_t fakedrop = FALSE;	/* simulate dropping of 1st packet */
boolean_t opt_confirm_receipt = FALSE;
port_t search_port;

/* publishing globals */
u_char 	mode;			/* broadcast, unicast or multicast	*/
boolean_t done;			/* finished this message		*/
u_char 	lingering = 0;		/* countdown until finished message	*/
publish_packet_t perm_packet;	/* permission packet			*/
u_long  cache_limit = DEF_CACHESIZE;	 /* storage limit for file data */

/* host and socket globals */
host_t	multicast_address;	/* this is determined by secretary 	*/
char	multicast_ttl;		/* this is determined by secretary 	*/
host_t	secretary_host = 0;	/* host address of the secretary	*/
int 	group_data_sock; 	/* UDP socket used for msg send/resend  */
port_t 	group_data_port;	/* port used for clients<-publisher     */
int	control_sock;		/* TCP socket for requesting resends 	*/
port_t	control_port;		/* port used for clients->publisher	*/
port_t 	publish_port; 		/* port used for publisher<->secretary	*/

/* file and packet globals */
msg_id_t msg_id;		/* message id assigned by secretary	*/
seq_t 	seq_total;		/* number of packets needed total 	*/
seq_t 	seq;			/* current packet number		*/
u_short infosize = sizeof(file_info_t);	/* size of file info header	*/
u_long  msg_size;		/* total amount of data to be sent 	*/
u_long 	fsize;			/* file size 				*/
u_short data_size;		/* file data part of general packet	*/
u_short last_data_size;		/* file data part of last data packet	*/
u_short	bytes_to_send;		/* size of current packet		*/

/* tuning globals */
long	forced_interval = 0;	/* if we want to override the secretary */
long	no_multicast = 0;	/* if we want to disable multicast      */
long	interval;		/* delay between each packet send	*/
long	min_interval = DEF_MIN_INTERVAL;	/* min between packets	*/
long	max_interval = DEF_MAX_INTERVAL;	/* max between packets	*/
u_int	stepsize = DEF_STEPSIZE;/* increment/decrement size of interval	*/
seq_t	window_size = DEF_WINDOWSIZE;		
boolean_t unicast_resends = FALSE;

/* packet size globals */
#ifdef MIPS
u_long	unicast_size   = (MAX_UNICAST_DATA_SIZE - 1000);
#else
u_long	unicast_size   = MAX_UNICAST_DATA_SIZE;	 /* unicast pkt size	*/
#endif
u_long	multicast_size = MAX_MULTICAST_DATA_SIZE;/* multicast pkt size	*/
u_long	broadcast_size = MAX_BROADCAST_DATA_SIZE;/* broadcast pkt size	*/

/* definition of our table of hosts in our group */
host_t	memberlist[MAXMEMBERS];	/* this is stored in network byte order	*/
u_short members = 0;		/* number of members currently in group	*/

/* statistics globals */
static u_long sends;		/* for double-checking stuff		*/
static u_long resends;		/* for measuring efficiency		*/
static struct timeval t_start, t_end;	/* start & end time		*/
static struct timezone tzp;	/* silly timezone variable		*/

int main(argc, argv)
int argc;
char *argv[];
{
	int c;
	extern char *optarg;
	extern int optind;

	/* not used by client, but needed for call to FindSecretary() */
	port_t 	admin_port;
	int	verflg = 0;

	search_port = htons(AFDP_SEARCH_PORT);  /* default value */
	progname = argv[0];
	setlog(progname, LOG_WARNING, LOG_STDERR, NULL, NULL);

	while ((c=getopt(argc,argv,"DVa:c:dg:ho:p:rs:C:FMNRSU:b:i:u:w:"))!= -1){
		switch (c) {
		/* standard afdp options first */
		case 'D':
			daemonize = TRUE;
			break;
		case 'V':
			verflg++;	/* print out version message */
			break;
	        case 'a':
#ifdef ACL
			ReadACLFile(optarg);
#else
			logmsg(LOG_ERR,"Access control lists not supported\n");
			exit(3);
#endif
			break;
		case 'c':
			command = optarg;
			break;
		case 'd':
			debug++;
			break;
		case 'g':
			group = atol(optarg);
			break;
		case 'h':
                    	Usage();	/* print out help message */
			break;
		case 'o':		/* afdp protocol options */
			if(strcmp(optarg, "confirm") == 0) {
				logmsg(LOG_DEBUG, "Option: confirm receipt");
				opt_confirm_receipt = TRUE;
			}
			else if(strcmp(optarg, "warn") == 0)
				logmsg(LOG_DEBUG, "Option: warn for resends");
			else
				logmsg(LOG_NOTICE,"BAD option: %s", optarg);
		case 'p':
			portname = optarg;
			/* logmsg(LOG_NOTICE,"portname is %s\n", portname); */
			break;
		case 'r':
			if(getuid()) {
			   logmsg(LOG_ERR,"must be root to use reserved ports");
			   exit(4);
			}
			reserved_ports = TRUE;
			portname = "afdp_res";
			search_port = htons(AFDP_RESERVED_PORT);
			break;
		case 's':
			secretary_host = HostnameToAddr(optarg);
			/* logmsg(LOG_NOTICE,"looking for secretary on %s\n",
				HostaddrToName(secretary_host)); */
			break;
		/* options for afdpsend, afdp client (publisher) */
		case 'C':
			cache_limit = atoi(optarg);
			break;
		case 'F':
			fakedrop = TRUE;
			break;
		case 'M':
			no_multicast = TRUE;
			break;
		case 'N':
			min_interval = (100 * mSEC);
			break;
		case 'R':
			min_interval = (2 * mSEC);
			break;
		case 'S':
			stepsize = (atoi(optarg))*mSEC;
			break;
		case 'U':
			unicast_resends = TRUE;
			break;
		case 'b':
			broadcast_size = atoi(optarg);
			break;
		case 'i':
			forced_interval = (atol(optarg))*mSEC;
			break;
		case 'u':
			multicast_size = unicast_size = atoi(optarg);
			break;
		case 'w':
			window_size = atoi(optarg);
			break;
               	default:
                    	Usage();
               	}
	}
	if ((optind == argc) && !verflg)
		Usage();

	if (verflg)
		Version();

	if (daemonize)
		setlog(progname, debug+LOG_WARNING, LOG_LOCAL7, LOG_PID, NULL);
	else
		setlog(progname, debug+LOG_WARNING, LOG_STDERR, NULL, NULL);

	if (debug)
		logmsg(LOG_NOTICE, "%s: debug level is %d\n", progname, debug);

	search_port = GetPort(portname, "udp", search_port);

	if (cache_limit < unicast_size) {
		logmsg(LOG_ERR,"Cache limit is too small.\n");
		logmsg(LOG_ERR,"Must be at least %ld bytes.\n", unicast_size);
		exit(4);
	}

	if (FindSecretary(group, &secretary_host, &admin_port,
			&publish_port, FALSE, search_port)) {
		logmsg(LOG_DEBUG, "secretary found on %s\n",
			HostaddrToName(secretary_host));
	}
	else {
		logmsg(LOG_ERR,"sorry, can't find secretary for group %lu.\n", group);
		exit(1);
	}

	for ( ; optind < argc; optind++) {
		if (access(argv[optind], 4))
			logmsg(LOG_ERR, "cannot access %s: %m", argv[optind]);
		else
			Publish(argv[optind]);
	}
	return(0);
}

void Usage()
{
  printf("usage: %s [-DVdhr] [-a aclfile] [-g group] [-p port] [-s secretary_host]",
	progname);
  printf(" [-c command] file ...\n");
  printf("protocol options:\n");
  printf("\t-o confirm\n");
  printf("\t-o warn\n");
  printf("tuning options:\n");
  printf("\t-C cache_size\n");
  printf("\t-F [fake drop of 1st packet]\n");
  printf("\t-M [disable multicasting]\n");
  printf("\t-N [nice mode]\n");
  printf("\t-R [rabid mode]\n");
  printf("\t-S stepsize\n");
  printf("\t-U [unicast resends]\n");
  printf("\t-b broadcast_size\n");
  printf("\t-i forced_interval\n");
  printf("\t-u unicast_size\n");
  printf("\t-w window_size\n");
  exit(2);
}

void Version()
{
	printf("This is %s, version %s ", progname, AFDP_VERSION);
#ifdef MULTICAST
	printf("[with multicast support].\n");
#else
	printf("[without multicast support].\n");
#endif
	exit(2);
}

int Publish(filename)
char *filename;
{
	int			tempsock, n, st;
	struct sockaddr_in	sin;
	char	myname[MAXHOSTNAMELEN];

	if ((fsize = OpenFile (filename)) == -1) {
		logmsg(LOG_ERR, "Publish: could not open file %s\n", filename);
		return(FAIL);
	}

	strcpy(myname, Hostname());
	msg_size = fsize + infosize;

	logmsg(LOG_NOTICE, "Publish: %s publishing file %s (%lu bytes) to group %lu\n",
		myname, filename, fsize, group);

	/* prepare a PUBLISH_REQUEST control packet */
	perm_packet.packet_type = PUBLISH_REQUEST;
	perm_packet.group_id = htonl( group );
	perm_packet.msg_size = htonl( msg_size );
#ifdef MULTICAST
	if (no_multicast)
		perm_packet.publish_mode = MODE_BROADCAST;
	else
		perm_packet.publish_mode = MODE_MULTICAST;
#else
	perm_packet.publish_mode = MODE_BROADCAST;
#endif

	/* send the packet, and get the secretary's reply */
	st = TCPexchange(secretary_host, &tempsock, publish_port,
		&perm_packet, sizeof(perm_packet));
	if (st == FAIL) {
		logmsg(LOG_ERR, "Publish: failed to send PUBLISH_REQUEST\n" );
        	return( FAIL );
	}
#ifdef not_IDENT
	logmsg(LOG_NOTICE, "IDENT: secretary is %s@%s\n",
		PeerUser(tempsock), HostaddrToName(PeerHost(tempsock)));
#endif

	/* check the reply packet type */
	if (perm_packet.packet_type != PUBLISH_PERMISSION) {
		logmsg(LOG_ERR, "Publish: expected packet type %d, got %d\n",
			PUBLISH_PERMISSION, perm_packet.packet_type);
		return( FAIL );
	}

	/* Check the publisher status, and retry if busy */
	switch (perm_packet.status) {
	case OK:
		logmsg(LOG_INFO,"OK received, sending file ... ");
		break;
	case BUSY:
		logmsg(LOG_NOTICE, "Publish:  BUSY received, waiting ...\n" );
		Usleep(ntohl(perm_packet.time));
		return (Publish(filename));
	case REJECT:
		logmsg(LOG_NOTICE, "Publish:  permission denied\n" );
		return( FAIL );
	default:
		logmsg(LOG_ERR, "Publish:  bad status, cannot continue\n" );
		return( FAIL );
	}

	msg_id = ntohs(perm_packet.msg_id);
	group_data_port = perm_packet.group_data_port;
	mode = perm_packet.publish_mode;
	multicast_address = perm_packet.multicast_address;
	multicast_ttl = perm_packet.multicast_ttl;
	logmsg(LOG_NOTICE,"Multicast address is %s, ttl is %d\n",
		HostaddrToName(multicast_address), multicast_ttl);
	members = ntohs(perm_packet.host_list_size);
	if (members != 0) {
		logmsg(LOG_NOTICE, "Fetching list of %u group members\n", members);
		n = ReadStream(tempsock, memberlist, members*sizeof(host_t));
		if (n != (members*sizeof(host_t)) ) {
			logmsg(LOG_ERR,"Error when reading members list!\n");
		}
	}
	close(tempsock);

	if (forced_interval)
		interval = forced_interval;
	else
		interval = ntohl(perm_packet.interval);
	logmsg(LOG_NOTICE, "using packet interval of %ld ms\n", interval/mSEC);

	CreateCache (msg_size);

	CreateSock(&group_data_sock, SOCK_DGRAM, SO_BROADCAST);
#ifdef MULTICAST
	if (multicast_ttl > 1) {
		if (setsockopt (group_data_sock, IPPROTO_IP, IP_MULTICAST_TTL,
			&multicast_ttl, sizeof(multicast_ttl)) < 0) {
			fatal("setsockopt IP_MULTICAST_TTL: %m");
		}
		logmsg(LOG_INFO,"Using Multicast ttl of %d\n", multicast_ttl);
	}
#endif
	CreateSock(&control_sock, SOCK_STREAM, 0);
	BindSock(control_sock, &control_port, &sin);
	ListenSock(control_sock);

	done = FALSE;
	StartTimer(msg_id, interval, SendMsgPacket, 0);

	while(done == FALSE) {
		CheckInputs();
		ClockTick();
	}

	logmsg(LOG_DEBUG, "finished with that message!!!!\n");
	DestroyCache();
	close(control_sock);
	close(group_data_sock);
	return(SUCCESS);
}

void SendMsgPacket(dummy)
u_long dummy;		/* dummy argument, because of how StartTimer() works */
{
	static boolean_t msg_in_progress = FALSE;
	static data_packet_t packet;
	static long transmit_time;
	char *msgp;
	int speedup;
	publish_packet_t reply;

	dummy = 0; /* just to make lint shut up! */
	speedup = stepsize / 10; /* how much to speed up with each packet */

	if (!msg_in_progress) {
		/* either we are about to start a new message,
		 * or we are about to get rid of a message that has been sent.
		 */
		if (lingering) {
			if (--lingering) {
				/* we still have to wait a bit longer */
				logmsg(LOG_INFO,"linger for %d more seconds ...\n",
					lingering);
				StartTimer (msg_id, (long) SECOND, 
						SendMsgPacket, 0);
				return;
			}
			/* we are done with this message */
			done = TRUE;

			transmit_time = t_end.tv_usec - t_start.tv_usec;
			if (t_end.tv_sec != t_start.tv_sec) {
				transmit_time +=
				(t_end.tv_sec - t_start.tv_sec) * SECOND;
			}

			reply.packet_type = PUBLISH_COMPLETE;
			reply.group_id = htonl(group);
			reply.msg_id = htons(msg_id);
			reply.status = DONE;
			reply.time = htonl(transmit_time);
			reply.publish_mode = mode;
			logmsg(LOG_INFO,"sending PUBLISH_COMPLETE to secretary\n");

       			TCPsend(secretary_host, publish_port,
				(char *)&reply, sizeof(reply)); 

			/* print out statistics */
			logmsg(LOG_NOTICE,
			"Sent %lu bytes in %ld ms, throughput = %ld KB/s\n",
				msg_size, transmit_time/mSEC,
				(long) ((float)msg_size /(transmit_time/mSEC)));

			logmsg(LOG_NOTICE,
			"Used %lu packets + %lu resends, efficiency = %d%%\n",
				sends, resends, 
				(int) ((seq_total*100)/(seq_total+resends)) );

			logmsg(LOG_NOTICE,
			"Average rate: 1 message packet every %ld ms.\n",
				(long) (transmit_time/(mSEC * seq_total)) );
			return;
		}
		else {
			/* we must be starting a new message */

			seq = 0;	/* reset packet seq. number */
			sends = 0;
			resends = 0;

			logmsg(LOG_NOTICE, "SendMsg: initializing for msg %u\n",
				ntohs(perm_packet.msg_id));
			if (mode == MODE_UNICAST) {
				data_size = unicast_size;
				msgp="will unicast this msg";
			}
#ifdef MULTICAST
			/* if we support multicast, and the secretary told
			 * us that all the subscribers support it too,
			 * and the user hasn't disabled multicast,
			 * we'll use multicast. otherwise, we'll broadcast
			 */
			else if ((mode == MODE_MULTICAST) && !no_multicast) {
				data_size = multicast_size;
				msgp="will multicast this msg";
			}
#endif
			else {
				data_size = broadcast_size;
				msgp="will broadcast this msg";
			}
			logmsg(LOG_NOTICE, "%s, interval at %ld ms.\n",
				msgp, interval/mSEC);

			seq_total = msg_size / data_size; /* # of packets */
			last_data_size = msg_size % data_size;
			if (last_data_size) 
				seq_total++;		/* partial packet */

			packet.packet_type = MSG_SEND;
			packet.group_id    = perm_packet.group_id;
			packet.msg_id      = perm_packet.msg_id;
			packet.time	   = perm_packet.time;
			packet.seq_total   = htonl(seq_total);
			packet.control_port = control_port;
			packet.options      = opt_confirm_receipt;

			logmsg(LOG_NOTICE, "starting msg %u (will use %lu packets)",
				msg_id, seq_total);
			logmsg(LOG_NOTICE, "file is %lu bytes, data_size is %u.\n",
				fsize, data_size);

			msg_in_progress = TRUE;
			done = FALSE;

			/* take a timestamp of the start time */
			if (gettimeofday(&t_start, &tzp) != 0)
				logmsg(LOG_ERR, "gettimeofday: %m");
		}
	}
	/* send out a packet */
	seq++;					/* set sequence number */
	sends++;
#ifdef DEBUG
      if (seq == 1 && fakedrop)
	packet.seq_number = htonl(2);		/* fake drop of 1st pkt */
      else
#endif
	packet.seq_number = htonl(seq);		/* copy it into packet */

	if (seq == seq_total) {
		/* the last msg, a partial packet */
		bytes_to_send = last_data_size;
		msg_in_progress = FALSE; /* for next time we're called */
		lingering = LINGER_TIME;
	}
	else {
		/* a full-size packet */
		bytes_to_send = data_size;
	}
	packet.data_size = htons(bytes_to_send);

	Fetch(packet.data, seq);

	/* calculate how far into the file the subscriber should write */
	packet.offset = htonl(seq == 1 ? 0 : (seq - 1) * data_size - infosize);
	logmsg(LOG_DEBUG, "for packet %lu, offset is %lu\n",
		seq, ntohl(packet.offset));

	/* calculate checksum for packet data */
	packet.checksum = htonl (Checksum (packet.data, bytes_to_send));

	/* have to add header and packet.data_size to 'bytes_to_send' */
	bytes_to_send += DATA_PACKET_NONDATA_SIZE;

	/* now, send the packet to all our group members */
	logmsg(LOG_INFO,"sending packet %lu of %lu (msg %u), interval is %ld ms.\n",
			seq, seq_total, msg_id, interval/mSEC);

#ifdef ACL
	{
	  struct sockaddr_in myname;
	  int len = sizeof(struct sockaddr_in);
	  bzero(&myname, sizeof(myname));
	  if (!getsockname(group_data_sock, (struct sockaddr *)&myname, &len)){
	    logmsg(LOG_ERR, "getsockname: %m");
          }
 	  GenAuth(&packet, bytes_to_send, &(myname.sin_addr)); 
        }
#endif

	SendToGroup(group_data_sock, group_data_port,
			(char *) &packet, bytes_to_send);

	interval -= speedup;
	if (interval < min_interval)
		interval = min_interval;	/* this is our lower bound! */

	if (seq == seq_total) {
		/* take a timestamp of the end time */
		if (gettimeofday(&t_end, &tzp) != 0)
			logmsg(LOG_ERR, "gettimeofday: %m");
	}
	if ((mode == MODE_UNICAST) && (members > 1))
		/* use a smaller interval, because we pause between packets */
		StartTimer (msg_id, interval/2, SendMsgPacket, 0);
	else
		StartTimer (msg_id, interval, SendMsgPacket, 0);
}

void CheckInputs() 
{
	struct timeval timeout;
	fd_set r, w, e;
	int msgsock;

	FD_ZERO(&r);
	FD_ZERO(&w);
	FD_ZERO(&e);
	FD_SET(control_sock, &r);
    
 	timeout.tv_sec =  0;
	timeout.tv_usec = 100;

	if (select(getdtablesize(), &r, &w, &e, &timeout) > 0) {
		if (FD_ISSET(control_sock, &r)) {
			logmsg(LOG_INFO,"CheckInputs: got control message\n");
			msgsock = accept (control_sock, 0, 0);
#ifdef not_IDENT
			logmsg(LOG_NOTICE,"IDENT: control msg from subscriber %s@%s\n",
				PeerUser(msgsock),
				HostaddrToName(PeerHost(msgsock)));
#endif
			if (msgsock < 0) {
				fatal("accept failed");
			}
			else
				HandleControlPacket(msgsock);
	
			close (msgsock);  /* tear down connection */
		}
	}
}

void HandleControlPacket(sock) 
int sock;
{
	int got;
	host_t host;
	control_packet_t data;
	struct sockaddr_in peer;

	msg_id_t mid;			/* temp for message id #     */
	seq_t i;			/* temp for sequence #	     */

	GetPeer(sock, &peer);
	logmsg(LOG_INFO,"HandleControlPacket: got control msg from %s (port %u)\n",
		HostaddrToName(peer.sin_addr.s_addr), ntohs(peer.sin_port));

	host = peer.sin_addr.s_addr;

	got = ReadStream (sock, (char *)&data, sizeof(control_packet_t)); 
	if (got != sizeof(control_packet_t)) {
		logmsg(LOG_ERR,"HandleControlPacket: only got %d bytes\n", got);
		return;
	}

	mid = ntohs(data.msg_id);
	if (mid != msg_id) {
		logmsg(LOG_ERR,"I don't have message %u any more!\n", mid);
		/* it might be nice if we sent them back an error msg packet */
		return;
	}

#ifdef GROUP_IN_DATA_PACKET
	if (ntohl(data.group_id) != group) {
		logmsg(LOG_ERR,"control packet for wrong group: %lu, expecting %lu\n",
				ntohl(data.group_id), group);
		return;
	}
#endif

	switch(data.packet_type) {

	case SLOWDOWN_REQUEST:
		interval += stepsize;
		logmsg(LOG_INFO, "SLOWDOWN_REQUEST, interval is now %ld ms\n",
					interval/mSEC);
		break;
	case RESEND_REQUEST:
		i = ntohl(data.seq_number_low);
		if (i > seq) {
			logmsg(LOG_INFO,"ignoring request for pkt %lu!\n", i);
			return;
		}
		interval += stepsize;
		if (interval > max_interval) /* this is our upper bound! */
			interval = max_interval;

		logmsg(LOG_INFO,
		"RESEND_REQUEST for packets %lu -> %lu, interval now %ld ms\n",
				i, ntohl(data.seq_number_high), interval/mSEC);

		if (seq >= window_size && i < seq - window_size) {
			logmsg(LOG_NOTICE,"requested packet %lu out of window\n", i);
			SkipPacket(host, seq - window_size);
			i = seq - window_size;
		}

		while (i <= ntohl (data.seq_number_high)) {
			if (i <= seq) {
				ResendPacket(host, i);
				Usleep(interval);
			}
			else {
				logmsg(LOG_INFO,"packets above %lu not yet sent!\n",
				i);
				break; /* no sense in processing the others */
			}
			i++;
		}
		if (lingering) {
			logmsg(LOG_INFO, "Restarting linger timer ...\n");
			lingering = LINGER_TIME;
		}
		break;
	case SHUTDOWN:
		logmsg(LOG_INFO,"Got SHUTDOWN message from secretary. Bye.\n");
		exit(0);
	default:
		/* detect impossible condition */
		/* could be caused by error in byte-swapping? */
		fatal("HandleControlPacket");
		break;
	}
}

void SendToGroup(sock, port, msg, msg_size)
int sock;
port_t port;
char *msg;
int msg_size;

{
	u_int i;

	if(mode == MODE_UNICAST) {
		for (i = 0 ; i < members ; i++) {
			if (memberlist[i] == 0) {
			    /* entry is NULLHOST */
			    logmsg(LOG_DEBUG,"skipping NULLHOST entry %u\n", i);
			    continue;
			}
			logmsg(LOG_DEBUG,"Unicasting a MSG_SEND packet to %s\n",
				HostaddrToName(memberlist[i]));
			UDPsend(memberlist[i], sock, port,
				msg, msg_size);
			if (members > 1)
			    /* read SGI manual on ftimer(), we may
			     * be asking for finer control than is possible
			     */
			    Usleep(1); 	/* give it a breather */
		}
	}
#ifdef MULTICAST
	else if (mode == MODE_MULTICAST) {
		logmsg(LOG_DEBUG,"MULTICASTing a MSG_SEND packet to %s\n",
			HostaddrToName(multicast_address));
		UDPsend(multicast_address, sock, port,
			msg, msg_size);
	}
#endif
	else {
		Broadcast(sock, port, msg, msg_size);
	}
}

void ResendPacket(host, i)
host_t host;
seq_t i;
{
	static data_packet_t packet;

#ifdef ACL
	bzero(&packet, sizeof(packet));
#endif
	packet.packet_type = MSG_RESEND;
	packet.group_id    = perm_packet.group_id;
	packet.msg_id      = perm_packet.msg_id;
	packet.seq_number  = htonl(i);
	packet.seq_total   = htonl(seq_total);
	packet.time        = perm_packet.time;
	packet.control_port = control_port;

	if (i == seq_total) {
		/* the last msg, a partial packet */
		bytes_to_send = last_data_size;
	}
	else {
		/* a full-size packet */
		bytes_to_send = data_size;
	}
	packet.data_size = htons(bytes_to_send);

	Fetch(packet.data, i);

	/* calculate how far into the file the subscriber should write */
	packet.offset = htonl(i == 1 ? 0 : (i - 1) * data_size - infosize);
	logmsg(LOG_DEBUG, "for packet %lu, offset is %lu\n",
		seq, ntohl(packet.offset));

	/* calculate checksum for packet data */
	packet.checksum = htonl (Checksum (packet.data, bytes_to_send));

	/* have to add header and packet.data_size to 'bytes_to_send' */
	bytes_to_send += DATA_PACKET_NONDATA_SIZE;

#ifdef ACL
	{
	  struct sockaddr_in myname;
	  int len = sizeof(struct sockaddr_in);
	  bzero(&myname, sizeof(myname));
	  if (!getsockname(group_data_sock, (struct sockaddr *)&myname, &len)){
	    fatal("getsockname");
          }
	  GenAuth(&packet, bytes_to_send, &(myname.sin_addr));
	}
#endif

	/* now, send the packet to the host that requested it */
	logmsg(LOG_INFO,"ResendPacket: sending %s packet %lu of %lu (msg_id %u)\n",
			HostaddrToName(host), i, seq_total, msg_id);
	if(unicast_resends)
		UDPsend(host, group_data_sock, group_data_port,
			(char *) &packet, bytes_to_send);
	else
		SendToGroup(group_data_sock, group_data_port,
			(char *) &packet, bytes_to_send);
	resends++;

	/* take another timestamp of the end time */
	if (gettimeofday(&t_end, &tzp) != 0)
		logmsg(LOG_ERR, "gettimeofday: %m");
}

void SkipPacket(host, i)
host_t host;
seq_t i;
{
	static data_packet_t packet;

	packet.packet_type = MSG_SKIP;
	packet.group_id    = perm_packet.group_id;
	packet.msg_id      = perm_packet.msg_id;
	packet.seq_number  = htonl(i);
	packet.seq_total   = htonl(seq_total);
	packet.time        = perm_packet.time;
	packet.control_port= control_port;
	packet.offset 	   = htons(0);
	packet.data_size   = htons(0); 
	bytes_to_send = DATA_PACKET_NONDATA_SIZE;

#ifdef ACL
	{
	  struct sockaddr_in myname;
	  int len = sizeof(struct sockaddr_in);
	  bzero(&myname, sizeof(myname));
	  if (!getsockname(group_data_sock, (struct sockaddr *)&myname, &len)){
	    logmsg(LOG_ERR, "getsockname");
          }
	  GenAuth(&packet, bytes_to_send, &(myname.sin_addr));
	}
#endif

	/* now, send the packet to the host that requested it */
	logmsg(LOG_NOTICE,"SkipPacket: sending %s packet %lu of %lu (msg_id %u)\n",
			HostaddrToName(host), i, seq_total, msg_id);
	UDPsend(host, group_data_sock, group_data_port,
			(char *) &packet, bytes_to_send);
}
